# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Store structures together with energies and forces for potential fitting applications.

Basic usage:

>>> pr = Project("training")
>>> container = pr.create.job.TrainingContainer("small_structures")

Let's make a structure and invent some forces

>>> structure = pr.create.structure.ase_bulk("Fe")
>>> forces = numpy.array([-1, 1, -1])
>>> container.add_structure(structure, energy=-1.234, forces=forces, identifier="Fe_bcc")

If you have a lot of precomputed structures you may also add them in bulk from a pandas DataFrame

>>> df = pandas.DataFrame({ "name": "Fe_bcc", "atoms": structure, "energy": -1.234, "forces": forces })
>>> container.include_dataset(df)

You can retrieve the full database with :method:`~.TrainingContainer.to_pandas()` like this

>>> container.to_pandas()
name    atoms   energy  forces  number_of_atoms
Fe_bcc  ...
"""

from typing import Callable, Dict, Any, Optional

from warnings import catch_warnings

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from ase.atoms import Atoms as ASEAtoms

from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage, StructurePlots
from pyiron_atomistics.atomistics.structure.neighbors import NeighborsTrajectory
from pyiron_base import GenericJob, DataContainer, deprecate


class TrainingContainer(GenericJob, HasStructure):
    """
    Stores ASE structures with energies and forces.
    """

    def __init__(self, project, job_name):
        super().__init__(project=project, job_name=job_name)
        self.__name__ = "TrainingContainer"
        self.__hdf_version__ = "0.3.0"
        self._container = TrainingStorage()

        self.input = DataContainer(
            {"save_neighbors": True, "num_neighbors": 12},
            table_name="parameters"
        )

    def include_job(self, job, iteration_step=-1):
        """
        Add structure, energy and forces from job.

        Args:
            job (:class:`.AtomisticGenericJob`): job to take structure from
            iteration_step (int, optional): if job has multiple steps, this
            selects which to add
        """
        self._container.include_job(job, iteration_step)

    def include_structure(
        self,
        structure,
        energy=None,
        name=None,
        **properties
    ):
        """
        Add new structure to structure list and save energy and forces with it.

        For consistency with the rest of pyiron, energy should be in units of eV
        and forces in eV/A, but no conversion is performed.

        Args:
            structure_or_job (:class:`~.Atoms`): structure to add
            energy (float): energy of the whole structure
            forces (Nx3 array of float, optional): per atom forces, where N is
                the number of atoms in the structure
            stress (6 array of float, optional): per structure stresses in voigt
                notation
            name (str, optional): name describing the structure
        """
        self._container.include_structure(structure, name=name, energy=energy,
                                          **properties)

    def add_structure(
        self, structure, energy, forces=None, stress=None, identifier=None, **arrays
    ):
        """
        Add new structure to structure list and save energy and forces with it.

        For consistency with the rest of pyiron, energy should be in units of eV and forces in eV/A, but no conversion
        is performed.

        Args:
            structure_or_job (:class:`~.Atoms`): structure to add
            energy (float): energy of the whole structure
            forces (Nx3 array of float, optional): per atom forces, where N is the number of atoms in the structure
            stress (6 array of float, optional): per structure stresses in voigt notation
            name (str, optional): name describing the structure
        """
        self._container.add_structure(
            structure, energy, identifier=identifier, forces=forces, stress=stress, **arrays
        )

    def include_dataset(self, dataset):
        """
        Add a pandas DataFrame to the saved structures.

        The dataframe should have the following columns:
            - name: human readable name of the structure
            - atoms(:class:`ase.Atoms`): the atomic structure
            - energy(float): energy of the whole structure
            - forces (Nx3 array of float): per atom forces, where N is the number of atoms in the structure
            - stress (6 array of float): per structure stress in voigt notation
        """
        self._container.include_dataset(dataset)

    def _get_structure(self, frame=-1, wrap_atoms=True):
        return self._container.get_structure(frame=frame, wrap_atoms=wrap_atoms)

    def _number_of_structures(self):
        return self._container.number_of_structures

    def get_neighbors(self, num_neighbors=None):
        """
        Calculate and add neighbor information in each structure.

        If input.save_neighbors is True the data is automatically added to the internal storage and will be saved
        together with the normal structure data.

        Args:
            num_neighbors (int, optional): Number of neighbors to collect, if not given use value from input

        Returns:
            NeighborsTrajectory: neighbor information
        """
        if num_neighbors is None:
            num_neighbors = self.input.num_neighbors
        n = NeighborsTrajectory(
            has_structure=self,
            store=self._container if self.input.save_neighbors else None,
            num_neighbors=num_neighbors,
        )
        n.compute_neighbors()
        return n

    def get_elements(self):
        """
        Return a list of chemical elements in the training set.

        Returns:
            :class:`list`: list of unique elements in the training set as strings of their standard abbreviations
        """
        return self._container.get_elements()

    def to_pandas(self):
        """
        Export list of structure to pandas table for external fitting codes.

        The table contains the following columns:
            - 'name': human-readable name of the structure
            - 'ase_atoms': the structure as a :class:`.Atoms` object
            - 'energy': the energy of the full structure
            - 'forces': the per atom forces as a :class:`numpy.ndarray`, shape Nx3
            - 'stress': the per structure stress as a :class:`numpy.ndarray`, shape 6
            - 'number_of_atoms': the number of atoms in the structure, N

        Returns:
            :class:`pandas.DataFrame`: collected structures
        """
        return self._container.to_pandas()

    def to_list(self, filter_function=None):
        """
        Returns the data as lists of pyiron structures, energies, forces, and the number of atoms

        Args:
            filter_function (function): Function applied to the dataset (which is a pandas DataFrame) to filter it

        Returns:
            tuple: list of structures, energies, forces, and the number of atoms
        """
        return self._container.to_list(filter_function)

    def to_dict(self):
        return self._container.to_dict()

    def write_input(self):
        pass

    def collect_output(self):
        pass

    def run_static(self):
        self.status.running = True
        if self.input.save_neighbors:
            self.get_neighbors()
            self.to_hdf()
        self.status.finished = True

    def run_if_interactive(self):
        if self.input.save_neighbors:
            self.get_neighbors()
        self.to_hdf()
        self.status.finished = True

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(self.project_hdf5)
        self._container.to_hdf(self.project_hdf5, "structures")

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        hdf_version = self.project_hdf5.get("HDF_VERSION", "0.1.0")
        if hdf_version == "0.1.0":
            table = pd.read_hdf(
                self.project_hdf5.file_name, self.name + "/output/structure_table"
            )
            self.include_dataset(table)
        else:
            self._container = TrainingStorage()
            self._container.from_hdf(self.project_hdf5, "structures")
            if hdf_version == "0.3.0":
                self.input.from_hdf(self.project_hdf5, "parameters")

    def sample(
        self, name: str, selector: Callable[[StructureStorage, int], bool],
        delete_existing_job: bool = False
    ) -> "TrainingContainer":
        """
        Create a new TrainingContainer with structures filtered by selector.

        `self` must have status `finished`.  `selector` is passed the underlying :class:`StructureStorage` of this
        container and the index of the structure and return a boolean whether to include the structure in the new
        container or not.  The new container is saved and run.

        Args:
            name (str): name of the new TrainingContainer
            selector (Callable[[StructureStorage, int], bool]): callable that selects structure to include
            delete_existing_job (bool): if job with name exist, remove it first

        Returns:
            :class:`.TrainingContainer`: new container with selected structures

        Raises:
            ValueError: if a job with the given `name` already exists.
        """
        if not self.status.finished:
            raise ValueError(f"Job must be finished, not '{self.status}'!")
        cont = self.project.create.job.TrainingContainer(name, delete_existing_job=delete_existing_job)
        if not cont.status.initialized:
            raise ValueError(f"Job '{name}' already exists with status: {cont.status}!")
        cont._container = self._container.sample(selector)
        cont.run()
        return cont

    @property
    def plot(self):
        """
        :class:`.TrainingPlots`: plotting interface
        """
        return self._container.plot

    def iter(self, *arrays, wrap_atoms=True):
        """
        Iterate over all structures in this object and all arrays that are defined

        Args:
            wrap_atoms (bool): True if the atoms are to be wrapped back into the unit cell; passed to
                               :meth:`.get_structure()`
            *arrays (str): name of arrays that should be iterated over

        Yields:
            :class:`pyiron_atomistics.atomistitcs.structure.atoms.Atoms`, arrays: every structure attached to the object and queried arrays
        """
        yield from self._container.iter(*arrays, wrap_atoms=wrap_atoms)


class TrainingPlots(StructurePlots):
    """
    Simple interface to plot various properties of the structures inside the given :class:`.TrainingContainer`.
    """

    def energy_volume(self, crystal_systems=False):
        """
        Plot volume vs. energy.

        Volume and energy are normalized per atom before plotting.

        Args:
            crystal_systems (bool): if True, plot & label structures of different crystal systems separately.

        Returns:
            DataFrame: contains atomic energy and volumes in the columns 'E' and 'V'; if `crystal_systems` is given,
                       also contain space groups and crystal systems of each structure
        """

        N = self._store.get_array("length")
        E = self._store.get_array("energy") / N
        C = self._store.get_array("cell")
        V = np.linalg.det(C) / N

        df = pd.DataFrame({"V": V, "E": E})

        if crystal_systems:
            spg = self._calc_spacegroups()
            df = df.join(spg)
            for cs, dd in df.groupby("crystal_system"):
                plt.scatter(dd.V, dd.E, label=cs)
            plt.legend()
        else:
            plt.scatter(df.V, df.E)
        plt.xlabel(r"Atomic Volume [$\AA^3$]")
        plt.ylabel(r"Atomic Energy [eV]")

        return df

    def energy_distance(self, num_neighbors=None):
        """
        Plot energy vs minimum nearest neighbor distance in the structure.

        Args:
            num_neighbors (int): maximum number of neighbors to caluclate, when 'distances' are not defined in storage
                                 the default is the value from the previous call or 36
        """

        N = self._store.get_array("length")
        E = self._store.get_array("energy") / N
        neigh = self._calc_neighbors(num_neighbors=num_neighbors)
        D = np.array([d.min() for d in np.split(neigh['distances'][:, 0], N.cumsum())[:-1]])
        plt.scatter(D, E, marker='.')

    def forces(self, axis: Optional[int] = None):
        """
        Plot a histogram of all forces.

        Args:
            axis (int, optional): plot only forces along this axis, if not given plot all forces
        """
        f = self._store.get_array("forces")
        if axis is not None:
            f = f[:, axis]
        else:
            f = f.ravel()
        plt.hist(f, bins=20)
        plt.xlabel(r"Force [eV/$\mathrm{\AA}$]")


class TrainingStorage(StructureStorage):
    def __init__(self):
        super().__init__()
        self.add_array("energy", dtype=np.float64, per="chunk", fill=np.nan)
        self._table_cache = None
        self.to_pandas()

    def to_pandas(self):
        """
        Export list of structure to pandas table for external fitting codes.

        The table contains the following columns:
            - 'name': human-readable name of the structure
            - 'ase_atoms': the structure as a :class:`.Atoms` object
            - 'energy': the energy of the full structure
            - 'forces': the per atom forces as a :class:`numpy.ndarray`, shape Nx3
            - 'stress': the per structure stress as a :class:`numpy.ndarray`, shape 6
            - 'number_of_atoms': the number of atoms in the structure, N

        Returns:
            :class:`pandas.DataFrame`: collected structures
        """
        if self._table_cache is None or len(self._table_cache) != len(self):
            self._table_cache = pd.DataFrame(
                {
                    "name": [self.get_array("identifier", i) for i in range(len(self))],
                    "atoms": [self.get_structure(i) for i in range(len(self))],
                    "energy": [self.get_array("energy", i) for i in range(len(self))],
                }
            )
            if self.has_array("forces"):
                self._table_cache["forces"] = [self.get_array("forces", i) for i in range(len(self))]
            if self.has_array("stress"):
                self._table_cache["stress"] = [self.get_array("stress", i) for i in range(len(self))]
            self._table_cache["number_of_atoms"] = [
                len(s) for s in self._table_cache.atoms
            ]
        return self._table_cache

    def include_job(self, job, iteration_step=-1):
        """
        Add structure, energy, forces and pressures from an inspected or loaded job.

        The job must be an atomistic job.

        Forces and stresses are only added if present in the output.

        Args:
            job (:class:`.JobPath`, :class:`.AtomisticGenericJob`): job (path) to take structure from
            iteration_step (int, optional): if job has multiple steps, this selects which to add
        """

        kwargs = {
                "energy": job["output/generic/energy_pot"][iteration_step],
        }
        ff = job["output/generic/forces"]
        if ff is not None:
            kwargs["forces"] = ff[iteration_step]

        # HACK: VASP work-around, current contents of pressures are meaningless, correct values are in
        # output/generic/stresses
        pp = job["output/generic/stresses"]
        if pp is None:
            pp = job["output/generic/pressures"]
        if pp is not None and len(pp) > 0:
            stress = np.asarray(pp[iteration_step])
            if stress.shape == (3, 3):
                stress = np.array(
                    [
                        stress[0, 0],
                        stress[1, 1],
                        stress[2, 2],
                        stress[1, 2],
                        stress[0, 2],
                        stress[0, 1],
                    ]
                )
            kwargs["stress"] = stress

        ii = job["output/generic/indices"]
        if ii is not None:
            indices = ii[iteration_step]
        else:
            indices = job["input/structure/indices"]
        species=np.asarray(job["input/structure/species"])
        cell = job["output/generic/cells"][iteration_step]
        positions = job["output/generic/positions"][iteration_step]
        pbc=job["input/structure/cell/pbc"]

        self.add_chunk(
                len(indices),
                identifier=job.name,
                symbols=species[indices],
                positions=positions,
                cell=[cell],
                pbc=[pbc],
                **kwargs
        )

    @deprecate("Use add_structure instead")
    def include_structure(
        self,
        structure,
        energy,
        name=None,
        **properties
    ):
        """
        Add new structure to structure list and save energy and forces with it.

        For consistency with the rest of pyiron, energy should be in units of eV and forces in eV/A, but no conversion
        is performed.

        Args:
            structure_or_job (:class:`~.Atoms`): structure to add
            energy (float): energy of the whole structure
            forces (Nx3 array of float, optional): per atom forces, where N is the number of atoms in the structure
            stress (6 array of float, optional): per structure stresses in voigt notation
            name (str, optional): name describing the structure
        """
        self.add_structure(structure, identifier=name, energy=energy,
                           **properties)

    def add_structure(
        self,
        structure: Atoms,
        energy,
        identifier=None,
        **arrays
    ) -> None:
        if "forces" in arrays and not self.has_array("forces"):
            self.add_array("forces", shape=(3,), dtype=np.float64, per="element",
                           fill=np.nan)
        if "stress" in arrays and not self.has_array("stress"):
            # save stress in voigt notation
            self.add_array("stress", shape=(6,), dtype=np.float64, per="chunk",
                           fill=np.nan)
        super().add_structure(structure, identifier=identifier, energy=energy,
                              **arrays)

    def include_dataset(self, dataset):
        """
        Add a pandas DataFrame to the saved structures.

        The dataframe should have the following columns:
            - name: human readable name of the structure
            - atoms(:class:`ase.Atoms`): the atomic structure
            - energy(float): energy of the whole structure
            - forces (Nx3 array of float): per atom forces, where N is the number of atoms in the structure
            - charges (Nx3 array of floats):
            - stress (6 array of float): per structure stress in voigt notation
        """
        if (
            "name" not in dataset.columns
            or "atoms" not in dataset.columns
            or "energy" not in dataset.columns
        ):
            raise ValueError(
                "At least columns 'name', 'atoms' and 'energy' must be present in dataset!"
            )
        for row in dataset.itertuples(index=False):
            kwargs = {}
            if hasattr(row, "forces"):
                kwargs["forces"] = row.forces
            if hasattr(row, "stress"):
                kwargs["stress"] = row.stress
            self.add_structure(
                row.atoms, energy=row.energy, identifier=row.name, **kwargs
            )

    def to_list(self, filter_function=None):
        """
        Returns the data as lists of pyiron structures, energies, forces, and the number of atoms

        Args:
            filter_function (function): Function applied to the dataset (which is a pandas DataFrame) to filter it

        Returns:
            tuple: list of structures, energies, forces, and the number of atoms
        """
        data_table = self.to_pandas()
        if filter_function is not None:
            data_table = filter_function(data_table)
        structure_list = data_table.atoms.to_list()
        energy_list = data_table.energy.to_list()
        if "forces" not in data_table.columns:
            raise ValueError("no forces defined in storage; call to_dict() instead.")
        force_list = data_table.forces.to_list()
        num_atoms_list = data_table.number_of_atoms.to_list()

        return (structure_list, energy_list, force_list, num_atoms_list)

    def to_dict(self) -> Dict[str, Any]:
        """Return a dictionary of all structures and training properties."""
        dict_arrays = {}

        # Get structure information.
        dict_arrays['structure'] = list(self.iter_structures())

        # Some arrays are only for internal usage or structure information that
        # was already saved in dict['structure'].
        internal_arrays = ['start_index', 'length', 'cell', 'pbc', 'positions',
                           'symbols']
        for array in self.list_arrays():
            # Skip internal arrays.
            if array in internal_arrays:
                continue

            dict_arrays[array] = self.get_array_ragged(array)
        return dict_arrays

    def iter(self, *arrays, wrap_atoms=True):
        """
        Iterate over all structures in this object and all arrays that are defined

        Args:
            wrap_atoms (bool): True if the atoms are to be wrapped back into the unit cell; passed to
                               :meth:`.get_structure()`
            *arrays (str): name of arrays that should be iterated over

        Yields:
            :class:`pyiron_atomistics.atomistitcs.structure.atoms.Atoms`, arrays: every structure attached to the object and queried arrays
        """
        array_vals = (self.get_array_ragged(a) for a in arrays)
        yield from zip(self.iter_structures(), *array_vals)

    @property
    def plot(self):
        """
        :class:`.TrainingPlots`: plotting interface
        """
        if self._plots is None:
            self._plots = TrainingPlots(self)
        return self._plots



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import os
from pyiron_atomistics.lammps.base import Input
from pyiron_atomistics.lammps.interactive import LammpsInteractive
from pyiron_atomistics.atomistics.structure.atoms import Atoms
from pyiron_atomistics.atomistics.structure.structurestorage import StructureStorage
from pyiron_contrib.atomistics.mlip.mlip import read_cgfs
from pyiron_contrib.atomistics.mlip.cfgs import loadcfgs
from pyiron_base import GenericParameters

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class LammpsMlip(LammpsInteractive):
    def __init__(self, project, job_name):
        super(LammpsMlip, self).__init__(project, job_name)
        self.input = MlipInput()
        self.__name__ = "LammpsMlip"
        self.__version__ = None  # Reset the version number to the executable is set automatically
        self._executable = None
        self._executable_activate()
        self._selected_structures = None

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        super(LammpsMlip, self).set_input_to_read_only()
        self.input.mlip.read_only = True

    def write_input(self):
        super(LammpsMlip, self).write_input()
        if self.input.mlip['mtp-filename'] == 'auto':
            self.input.mlip['mtp-filename'] = os.path.basename(self.potential['Filename'][0][0])
        self.input.mlip.write_file(file_name="mlip.ini", cwd=self.working_directory)

    def convergence_check(self):
        for line in self["error.out"]:
            if line.startswith("MLIP: Breaking threshold exceeded"): return False
        return True

    def enable_active_learning(self, threshold=2.0, threshold_break=5.0):
        """
        Enable active learning during MD run.

        Automatically collect structures on which the potential is extrapolating.

        Args:
            threshold (float): select structures with extrapolation grade larger than this
            threshold_break (float): stop the MD run after seeing a structure with extrapolation grade larger than this
        """
        self.executable.accepted_return_codes += [8]
        self.input.mlip.load_string(f"""\
mtp-filename auto
calculate-efs TRUE
select TRUE
select:threshold {threshold}
select:threshold-break {threshold_break}
select:save-selected selected.cfg
select:load-state state.mvs
select:log selection.log
write-cfgs:skip 0
""")

    def _get_selection_file(self):
        return os.path.join(self.working_directory, self.input.mlip['select:save-selected'])

    @property
    def _selection_enabled(self):
        return self.input.mlip["select"] == "TRUE"

    @property
    def selected_structures(self):
        """
        :class:`.StructureStorage`: structures that the potential extrapolated on during the run.

        Only available if :method:`.enable_active_learning` was called and once the job has been collected.
        """
        if not (self.status.collect or self.status.finished or self.status.not_converged):
            raise ValueError("Selected structures are only available once the job has finished!")
        if not self._selection_enabled:
            raise ValueError("Selected structures are only available after calling enable_active_learning()!")
        if self._selected_structures is None:
            if "selected" in self["output"].list_groups():
                self._selected_structures = self["output/selected"].to_object()
            else:
                self._selected_structures = StructureStorage()
        return self._selected_structures

    def collect_output(self):
        super(LammpsMlip, self).collect_output()
        if 'select:save-selected' in self.input.mlip._dataset['Parameter']:
            file_name = self._get_selection_file()
            if os.path.exists(file_name):
                for cfg in loadcfgs(file_name):
                    self.selected_structures.add_structure(
                            Atoms(species=self.structure.species, indices=cfg.types, positions=cfg.pos, cell=cfg.lat,
                                  pbc=[True, True, True]),
                            mv_grade=cfg.grade
                    )
                cell, positions, forces, stress, energy, indicies, grades, jobids, timesteps = read_cgfs(file_name=file_name)
                with self.project_hdf5.open("output/mlip") as hdf5_output:
                    hdf5_output['forces'] = forces
                    hdf5_output['energy_tot'] = energy
                    hdf5_output['pressures'] = stress
                    hdf5_output['cells'] = cell
                    hdf5_output['positions'] = positions
                    hdf5_output['indicies'] = indicies
            self.selected_structures.to_hdf(self.project_hdf5.open("output"), "selected")


class MlipInput(Input):
    def __init__(self):
        self.mlip = MlipParameter()
        super(MlipInput, self).__init__()

    def to_hdf(self, hdf5):
        """
        
        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.to_hdf(hdf5_input)
        super(MlipInput, self).to_hdf(hdf5)

    def from_hdf(self, hdf5):
        """
        
        Args:
            hdf5:
        Returns:
        """
        with hdf5.open("input") as hdf5_input:
            self.mlip.from_hdf(hdf5_input)
        super(MlipInput, self).from_hdf(hdf5)


class MlipParameter(GenericParameters):
    def __init__(self, separator_char=' ', comment_char='#', table_name="mlip_inp"):
        super(MlipParameter, self).__init__(separator_char=separator_char, comment_char=comment_char, table_name=table_name)
        
    def load_default(self, file_content=None):
        if file_content is None:
            file_content = '''\
mtp-filename auto
select FALSE
'''
        self.load_string(file_content)




# encoding: utf-8
# Copyright (c) Georg-August-Universität Göttingen - Behler Group
# Distributed under the terms of "New BSD License", see the LICENSE file.
"""Utility functions for use with the pyiron Hamiltonian of RuNNer.

.. _RuNNer online documentation:
   https://theochem.gitlab.io/runner
"""

from typing import List

import numpy as np

from ase.atoms import Atoms

from pyiron_atomistics.atomistics.structure.atoms import pyiron_to_ase

from runnerase.singlepoint import RunnerSinglePointCalculator

from ..atomistics.job.trainingcontainer import TrainingContainer


def container_to_ase(container: TrainingContainer) -> List[Atoms]:
    """Convert a `TrainingContainer` into a list of ASE Atoms objects.

    Args:
        container (TrainingContainer): The training data to be converted.

    Returns:
        structures (List[Atoms]): A list of ASE Atoms objects.
    """
    structure_lst = []

    arrays = container.to_dict()
    arraynames = arrays.keys()

    # Iterate over the structures by zipping the dictionary values.
    for properties in zip(*arrays.values()):
        zipped = dict(zip(arraynames, properties))

        # Retrieve atomic positions, cell vectors, etc.
        atoms = pyiron_to_ase(zipped['structure'])

        # Attach charges to the Atoms object.
        if 'charges' in zipped:
            atoms.set_initial_charges(zipped['charges'])

        # Store all properties that will be saved on the calculator below.
        calc_properties = {'energy': None, 'forces': None, 'totalcharge': None}
        for prop in calc_properties:
            if prop in zipped:
                calc_properties[prop] = zipped[prop]

        # Overwrite the totalcharge if the property was not present.
        if calc_properties['totalcharge'] is None:
            totalcharge = np.sum(atoms.get_initial_charges())
            calc_properties['totalcharge'] = totalcharge

        # Storage energies, forces, and totalcharge on a calculator object.
        atoms.calc = RunnerSinglePointCalculator(
            atoms=atoms,
            energy=calc_properties['energy'],
            forces=calc_properties['forces'],
            totalcharge=calc_properties['totalcharge']
        )

        structure_lst.append(atoms)

    return structure_lst


def ase_to_container(
    structures: List[Atoms],
    container: TrainingContainer
) -> None:
    """Append `structures` to `TrainingContainer`.

    Args:
        structures (List[Atoms]): A list of ASE Atoms objects to be stored on
            the given `container`.
        container (TrainingContainer): The container to which the data will be
            appended.
    """
    for structure in structures:

        properties = {}

        # If the structure has a calculator attached to it, get energies,
        # forces, etc.
        if structure.calc is not None:
            properties['energy'] = structure.get_potential_energy()
            properties['forces'] = structure.get_forces()
            properties['charges'] = structure.get_initial_charges()

            if isinstance(structure.calc, RunnerSinglePointCalculator):
                properties['totalcharge'] = structure.calc.get_property(
                    'totalcharge'
                )

            else:
                properties['totalcharge'] = np.sum(properties['charges'])

        # StructureStorage needs a `spins` property, but not all ASE Atoms
        # objects have that.
        if not hasattr(structure, 'spins'):
            structure.spins = None

        container.include_structure(structure, **properties)


def pad(array: np.ndarray, desired_length: int) -> np.ndarray:
    """Pad `array` with `np.NaN` rows up to `desired_length`.

    This routine pads an array of symmetry function values with np.NaN in those
    places where the index (first column of sfvalue arrays) is missing.

    Args:
        array (np.ndarray): The array to be padded. The first column must
            contain a continuous index.
        desired_length (int): The final desired length of the array.

    Returns:
        array_padded (np.ndarray): The padded array.
    """
    # Create a sequence of missing indices.
    all_indices = np.arange(0, desired_length, 1)
    contained_indices = array[:, 0].astype(int)
    missing_indices = np.delete(all_indices, contained_indices)

    # Create an np.NaN-filled array for the padded sfvalues data.
    array_padded = np.empty(shape=(desired_length, array.shape[1]))
    array_padded[:] = np.NaN

    # Insert first the data for this atom, second the np.NaN values.
    array_padded[:len(contained_indices), :] = array
    array_padded[len(contained_indices):, 0] = missing_indices

    return array_padded


def unpad(array: np.ndarray):
    """Remove all rows containing NaN values from an ndarray."""
    return array[~np.isnan(array).any(axis=1), :]



"""
This module provides helper functions for evaluating data relative to type hints, and
type hints relative to each other.
"""

import types
import typing
from collections.abc import Callable

from typeguard import check_type, TypeCheckError


def valid_value(value, type_hint) -> bool:
    try:
        return isinstance(value, type_hint)
    except TypeError:
        # Subscripted generics cannot be used with class and instance checks
        try:
            # typeguard handles this case
            check_type(value, type_hint)
            return True
        except TypeCheckError:
            # typeguard raises an error on a failed check
            return False


def type_hint_to_tuple(type_hint) -> tuple:
    if isinstance(type_hint, (types.UnionType, typing._UnionGenericAlias)):
        return typing.get_args(type_hint)
    else:
        return type_hint,


def type_hint_is_as_or_more_specific_than(hint, other) -> bool:
    hint_origin = typing.get_origin(hint)
    other_origin = typing.get_origin(other)
    if set([hint_origin, other_origin]) & set([types.UnionType, typing.Union]):
        # If either hint is a union, turn both into tuples and call recursively
        return all(
            [
                any(
                    [
                        type_hint_is_as_or_more_specific_than(h, o)
                        for o in type_hint_to_tuple(other)
                    ]
                )
                for h in type_hint_to_tuple(hint)
            ]
        )
    elif hint_origin is None and other_origin is None:
        # Once both are raw classes, just do a subclass test
        try:
            return issubclass(hint, other)
        except TypeError:
            return hint == other
    elif hint_origin == other_origin:
        # If they both have an origin, break into arguments and treat cases
        hint_args = typing.get_args(hint)
        other_args = typing.get_args(other)
        if len(hint_args) == 0 and len(other_args) > 0:
            # Failing to specify anything is not being more specific
            return False
        elif hint_origin in [dict, tuple, Callable]:
            # for these origins the order of arguments matters
            if len(other_args) == 0:
                # If the other doesn't specify _any_ arguments, we must be more specific
                return True
            elif len(other_args) == len(hint_args):
                # If they both specify arguments, they should be more specific 1:1
                return all(
                    [
                        type_hint_is_as_or_more_specific_than(h, o)
                        for o, h in zip(other_args, hint_args)
                    ]
                )
            else:
                # Otherwise they both specify but a mis-matching number of args
                return False
        else:
            # Otherwise order doesn't matter so make sure the arguments are a subset
            return all(
                [
                    any(
                        [
                            type_hint_is_as_or_more_specific_than(h, o)
                            for o in other_args
                        ]
                    )
                    for h in hint_args
                ]
            )
    else:
        # Lastly, if they both have origins, but different ones, fail
        return False


# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

"""
Job class for calculating the MTP descriptors for a set of structures.
"""

__author__ = "Marvin Poul"
__copyright__ = "Copyright 2021, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Marvin Poul"
__email__ = "poul@mpie.de"
__status__ = "development"
__date__ = "Mar 25, 2021"

from pyiron_base import Settings, DataContainer, GenericJob, Executable, FlattenedStorage
from pyiron_atomistics.atomistics.structure.has_structure import HasStructure
from .cfgs import Cfg, savecfgs

import os.path
import shutil

import numpy as np

# This class expects the job executable to read the potential and configurations and write the descriptors to a few
# hard-coded paths
_POTENTIAL_PATH="potential.mtp"
_INPUT_PATH="input.cfg"
_OUTPUT_PATH="out.xyz"

class MlipDescriptors(GenericJob):
    """
    Calculates the descriptors for a set of structures from a given MTP potential.

    As input a :class:`.Mlip` job with the MTP potential and a
    :class:`pyiron_atomistics.atomistics.job.structurecontainer.StructureContainer` with the structures of interest
    must be set to the respective attributes of the job :attribute:`.potential` and :attribute:`.structures`.  There are
    no other input parameters.
    """
    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.input = DataContainer({'potential_job_id': None, 'structure_container_id': None},
                                   table_name="parameters")
        self._descriptors = FlattenedStorage()
        self._executable_activate()

    @property
    def potential(self):
        """
        :class:`.Mlip`: job that contains a fitted MTP potential or `None` if not set
        """
        return self.project.load(self.input.potential_job_id)

    @potential.setter
    def potential(self, job):
        self.input.potential_job_id = job.id

    @property
    def structures(self):
        """
        :class:`pyiron_atomistics.atomistics.structure.has_structure.HasStructure`:
                structure container that keeps the structures to be evaluated
        """
        return self.project.load(self.input.structure_container_id)

    @potential.setter
    def structures(self, job: HasStructure):
        if not isinstance(job, HasStructure):
            raise TypeError("job class must derive from HasStructure")
        self.input.structure_container_id = job.id

    def write_input(self):
        self._create_working_directory()

        potential = self.project.load(self.input.potential_job_id)
        shutil.copy2(potential.potential_files[0],
                     os.path.join(self.working_directory, _POTENTIAL_PATH))

        cfgs = []
        container = self.project.load(self.input.structure_container_id)
        for structure in container.iter_structures():
            c = Cfg()
            c.pos = structure.positions
            c.lat = structure.cell
            c.types = structure.indices
            cfgs.append(c)

        savecfgs(filename=os.path.join(self.working_directory, _INPUT_PATH), cfgs=cfgs)

    def collect_output(self):

        def parse(f):
            s = f.readline()
            while s != "":
                N = int(s)
                M = int(f.readline().split()[-1])
                x = np.empty((N,M))
                for i, l in zip(range(N), f):
                    d = np.fromiter(map(float, l.split()), dtype=float)
                    x[i, :] = d[4:4 + M]
                yield x
                s = f.readline()

        file_name = os.path.join(self.working_directory, _OUTPUT_PATH)
        self._descriptors = FlattenedStorage()
        with open(file_name) as f:
            for chunk in parse(f):
                self._descriptors.add_chunk(chunk.shape[0], descriptors=chunk)
        with self.project_hdf5.open("output") as hdf:
            self._descriptors.to_hdf(hdf=hdf, group_name="descriptors")

    def to_hdf(self, hdf=None, group_name=None):
        super().to_hdf(hdf=hdf, group_name=group_name)
        self.input.to_hdf(hdf=self.project_hdf5)

    def from_hdf(self, hdf=None, group_name=None):
        super().from_hdf(hdf=hdf, group_name=group_name)
        self.input.from_hdf(hdf=self.project_hdf5)
        if self.status.finished:
            with self.project_hdf5.open("output") as hdf:
                self._descriptors.from_hdf(hdf=hdf, group_name="descriptors")



from pyiron_atomistics.atomistics.job.sqs import SQSJob, get_sqs_structures
from pyiron_base import state


class SQSJobWithoutOutput(SQSJob):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super(SQSJobWithoutOutput, self).__init__(project, job_name)
        self._interactive_disable_log_file = True

    def to_hdf(self, hdf=None, group_name=None):
        """

        Args:
            hdf:
            group_name:

        Returns:

        """
        if not self._interactive_disable_log_file:
            super(SQSJobWithoutOutput, self).to_hdf(hdf=hdf, group_name=group_name)

    def run_static(self):
        if not self._interactive_disable_log_file:
            super(SQSJobWithoutOutput, self).run_static()
        else:
            self._lst_of_struct, decmp, iterations, cycle_time = get_sqs_structures(
                structure=self.structure,
                mole_fractions={k: v for k, v in self.input.mole_fractions.items()},
                weights=self.input.weights,
                objective=self.input.objective,
                iterations=self.input.iterations,
                output_structures=self.input.n_output_structures,
                num_threads=self.server.cores,
            )

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(SQSJobWithoutOutput, self).refresh_job_status()



from __future__ import print_function
# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from logging import getLogger
from inspect import getargspec
from pydoc import locate
from itertools import islice
import re
"""
Classes for handling protocols, particularly setting up input and output pipes.
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "18 July, 2019"



def ordered_dict_get_index(ordered_dict, index):
    """
    Gets the object at "index" of an collections.OrderedDict without copying the keys list
    Args:
        ordered_dict: (collections.OrderedDict) the dict to get the value from
        index: (int) the index

    Returns: (object) the object at "index"

    """
    return ordered_dict[next(islice(ordered_dict, index, None))]

def ordered_dict_get_last(ordered_dict):
    """
    Gets the last most recently added object of an collections.OrderedDict instance

    Args:
        ordered_dict: (collections.OrderedDict) the dict to get the value from

    Returns: (object) the object at the back

    """

    return ordered_dict[next(reversed(ordered_dict))]


class LoggerMixin(object):
    """
    A class which is meant to be inherited from. Provides a logger attribute. The loggers name is the fully
    qualified type name of the instance
    """

    def fullname(self):
        """
        Returns the fully qualified type name of the instance

        Returns:
            str: fully qualified type name of the instance
        """
        return '{}.{}'.format(self.__class__.__module__, self.__class__.__name__)

    @property
    def logger(self):
        return getLogger(self.fullname())


def requires_arguments(func):
    """
    Determines if a function of method needs arguments, ingores self

    Args:
        func: (callable) the callable

    Returns: (bool) wether arguments (except "self" for methods) are needed

    """
    args, varargs, varkw, defaults = getargspec(func)
    if defaults:
        args = args[:-len(defaults)]
    # It could be a bound method too
    if 'self' in args:
        args.remove('self')
    return len(args) > 0

flatten = lambda l: [item for sublist in l for item in sublist]

def fullname(obj):
    """
    Returns the fully qualified class name of an object

    Args:
        obj: (object) the object

    Returns: (str) the class name

    """
    obj_type = type(obj)
    return '{}.{}'.format(obj_type.__module__, obj_type.__name__)


def get_cls(string):
    return locate([v for v in re.findall(r'(?!\.)[\w\.]+(?!\.)', string)if v != 'class'][0])


def is_iterable(o):
    """
    Convenience method to test for an iterator

    Args:
        o: the object to test

    Returns:
        bool: wether the input argument is iterable or not
    """
    try:
        iter(o)
    except TypeError:
        return False
    else:
        return not isinstance(o, str)

# convenience function to ensure the passed argument is iterable
ensure_iterable = lambda v: v if is_iterable(v) else [v]


class Registry(type):
    def __init__(cls, name, bases, nmspc):
        super(Registry, cls).__init__(name, bases, nmspc)
        if not hasattr(cls, 'registry'):
            cls.registry = set()
        cls.registry.add(cls)
        cls.registry -= set(bases) # Remove base classes


from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from warnings import warn

from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.type_hinting import (
    valid_value, type_hint_is_as_or_more_specific_than
)

if typing.TYPE_CHECKING:
    from pyiron_contrib.workflow.node import Node


class Channel(HasToDict, ABC):
    """
    Channels facilitate the flow of information (data or control signals) into and
    out of nodes.
    They have a label and belong to a node.

    Input/output channels can be (dis)connected from other output/input channels, and
    store all of their current connections in a list.
    """

    def __init__(
            self,
            label: str,
            node: Node,
    ):
        self.label = label
        self.node = node
        self.connections = []

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def connect(self, *others: Channel):
        pass

    def disconnect(self, *others: Channel):
        for other in others:
            if other in self.connections:
                self.connections.remove(other)
                other.disconnect(self)

    def disconnect_all(self):
        self.disconnect(*self.connections)

    @property
    def connected(self):
        return len(self.connections) > 0

    def _already_connected(self, other: Channel):
        return other in self.connections

    def __iter__(self):
        return self.connections.__iter__()

    def __len__(self):
        return len(self.connections)

    def to_dict(self):
        return {
            "label": self.label,
            "connected": self.connected,
            "connections": [f"{c.node.label}.{c.label}" for c in self.connections]
        }


class DataChannel(Channel, ABC):
    """
    Data channels control the flow of data on the graph.
    They store this data in a `value` attribute.
    They may optionally have a type hint.
    They have a `ready` attribute which tells whether their value matches their type
    hint.
    They may optionally have a storage priority (but this doesn't do anything yet).
    (In the future they may optionally have an ontological type.)

    The `value` held by a channel can be manually assigned, but should normally be set
    by the `update` method.
    In neither case is the type hint strictly enforced.
    Input channels will then propagate their value along to their owning node.
    Output channels with then propagate their value to all the input channels they're
    connected to.

    Type hinting is strictly enforced in one situation: when making connections to
    other channels and at least one data channel has a non-None value for its type hint.
    In this case, we insist that the output type hint be _as or more more specific_ than
    the input type hint, to ensure that the input always receives output of a type it
    expects. This behaviour can be disabled and all connections allowed by setting
    `strict_connections = False` on the relevant input channel.

    For simple type hints like `int` or `str`, type hint comparison is trivial.
    However, some hints take arguments, e.g. `dict[str, int]` to specify key and value
    types; `tuple[int, int, str]` to specify a tuple with certain values;
    `typing.Literal['a', 'b', 'c']` to specify particular choices;
    `typing.Callable[[float, float], str]` to specify a callable that takes particular
    argument types and has a return type; etc.
    For hints with the origin `dict`, `tuple`, and `typing.Callable`, the two hints must
    have _exactly the same arguments_ for one two qualify as "as or more specific".
    E.g. `tuple[int, int|float]` is as or more specific than
    `tuple[int|float, int|float]`, but not `tuple[int, int|float, str]`.
    For _all other hints_, we demand that the output hint arguments be a _subset_ of
    the input.
    E.g. `Literal[1, 2]` is as or more specific that both `Literal[1, 2]` and
    `Literal[1, 2, "three"]`.

    Warning:
        Type hinting in python is quite complex, and determining when a hint is
        "more specific" can be tricky. For instance, in python 3.11 you can now type
        hint a tuple with a mixture of fixed elements of fixed type, followed by an
        arbitrary elements of arbitrary type. This and other complex scenarios are not
        yet included in our test suite and behaviour is not guaranteed.

    TODO:
        In direct relation to the above warning, it may be nice to add a flag to
        channels to turn on/off the strict enforcement of type hints when making
        connections.
    """
    def __init__(
            self,
            label: str,
            node: Node,
            default: typing.Optional[typing.Any] = None,
            type_hint: typing.Optional[typing.Any] = None,
            storage_priority: int = 0,
            strict_connections: bool = True,
    ):
        super().__init__(label=label, node=node)
        self.default = default
        self.value = default
        self.type_hint = type_hint
        self.storage_priority = storage_priority
        self.strict_connections = strict_connections

    @property
    def ready(self):
        if self.type_hint is not None:
            return valid_value(self.value, self.type_hint)
        else:
            return True

    def connect(self, *others: DataChannel):
        for other in others:
            if self._valid_connection(other):
                self.connections.append(other)
                other.connections.append(self)
                out, inp = self._figure_out_who_is_who(other)
                inp.update(out.value)
            else:
                if isinstance(other, DataChannel):
                    warn(
                        f"{self.label} ({self.__class__.__name__}) and {other.label} "
                        f"({other.__class__.__name__}) were not a valid connection"
                    )
                else:
                    raise TypeError(
                        f"Can only connect two channels, but {self.label} "
                        f"({self.__class__.__name__}) got a {other} ({type(other)})"
                    )

    def _valid_connection(self, other):
        if self._is_IO_pair(other) and not self._already_connected(other):
            if self._both_typed(other):
                out, inp = self._figure_out_who_is_who(other)
                if not inp.strict_connections:
                    return True
                else:
                    return type_hint_is_as_or_more_specific_than(
                        out.type_hint, inp.type_hint
                    )
            else:
                # If either is untyped, don't do type checking
                return True
        else:
            return False

    def _is_IO_pair(self, other: DataChannel):
        return isinstance(other, DataChannel) and not isinstance(other, self.__class__)

    def _both_typed(self, other: DataChannel):
        return self.type_hint is not None and other.type_hint is not None

    def _figure_out_who_is_who(self, other: DataChannel) -> (OutputData, InputData):
        return (self, other) if isinstance(self, OutputData) else (other, self)

    def __str__(self):
        return str(self.value)

    def to_dict(self):
        d = super().to_dict()
        d["value"] = self.value
        d["ready"] = self.ready


class InputData(DataChannel):
    def update(self, value):
        self.value = value
        self.node.update()

    def activate_strict_connections(self):
        self.strict_connections = True

    def deactivate_strict_connections(self):
        self.strict_connections = False


class OutputData(DataChannel):
    def update(self, value):
        self.value = value
        for inp in self.connections:
            inp.update(self.value)


class SignalChannel(Channel, ABC):
    """
    Signal channels give the option control execution flow by triggering callback
    functions.

    Output channels can be called to trigger the callback functions of all input
    channels to which they are connected.
    """

    @abstractmethod
    def __call__(self):
        pass

    def connect(self, *others: Channel):
        for other in others:
            if self._valid_connection(other):
                self.connections.append(other)
                other.connections.append(self)
            else:
                if isinstance(other, SignalChannel):
                    warn(
                        f"{self.label} ({self.__class__.__name__}) and {other.label} "
                        f"({other.__class__.__name__}) were not a valid connection"
                    )
                else:
                    raise TypeError(
                        f"Can only connect two signal channels, but {self.label} "
                        f"({self.__class__.__name__}) got a {other} ({type(other)})"
                    )

    def _valid_connection(self, other) -> bool:
        return self._is_IO_pair(other) and not self._already_connected(other)

    def _is_IO_pair(self, other) -> bool:
        return isinstance(other, SignalChannel) \
            and not isinstance(other, self.__class__)


class InputSignal(SignalChannel):
    def __init__(
            self,
            label: str,
            node: Node,
            callback: callable,
    ):
        super().__init__(label=label, node=node)
        self.callback: callable = callback

    def __call__(self):
        self.callback()

    def __str__(self):
        return f"{self.label} runs {self.callback.__name__}"

    def to_dict(self):
        d = super().to_dict()
        d["callback"] = self.callback.__name__
        return d


class OutputSignal(SignalChannel):
    def __call__(self):
        for c in self.connections:
            c()

    def __str__(self):
        return f"{self.label} activates " \
               f"{[f'{c.node.label}.{c.label}' for c in self.connections]}"



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron_contrib.protocol.utils.misc import LoggerMixin, ensure_iterable, Registry
from pyiron_atomistics.atomistics.structure.atoms import Atoms
import numpy as np
import logging

"""
Classes to override compare
"""

__author__ = "Dominik Gehringer, Liam Huber"
__copyright__ = "Copyright 2019, Max-Planck-Institut für Eisenforschung GmbH " \
                "- Computational Materials Design (CM) Department"
__version__ = "0.0"
__maintainer__ = "Liam Huber"
__email__ = "huber@mpie.de"
__status__ = "development"
__date__ = "December 10, 2019"

try:
    from xxhash import xxh64_hexdigest
    hashfunction = xxh64_hexdigest
except ImportError:
    from hashlib import sha1
    logging.getLogger('pyiron_contrib.protocol.generic').debug('Falling back to SHA1 hashing')
    hashfunction = sha1

ensure_iterable_tuple = lambda o: tuple(ensure_iterable)


class Comparer(LoggerMixin, metaclass=Registry):
    """
    Class is aware of its subclasses. Subclasses must have a `type` attribute of type `type`.
    Compares two objects, where the behaviour can be overridden, and automatically determines the type.
    """

    def __init__(self, obj):
        super(Comparer, self).__init__()
        self._object = obj
        self._cls = type(obj)
        # this is a private object, one ought not access it
        self.__registry_cache = {}

        if not isinstance(self._object, self._cls):
            raise TypeError

    @property
    def object(self):
        return self._object

    def compatible_types(self, a, b):
        """
        Tests whether an equality comparison can be made between objects of type `a` and `b` by checking that these are
        the same type, or both either int or float.

        Args:
            a (type): The first type to be tested.
            b (type): The second type to be tested.

        Returns:
            (bool): True when the two types are the same or both belong to int and float.
        """
        if a != b and not self.both_are_int_or_float(a, b):
            return False
        else:
            return True

    @staticmethod
    def both_are_int_or_float(a, b):
        valid_list = [int, float]
        return a in valid_list and b in valid_list

    def _equals(self, b):
        if isinstance(b, Comparer):
            if not self.compatible_types(b._cls, self._cls):
                return False
            else:
                b = b._object
        elif not self.compatible_types(type(b), self._cls):
            self.logger.warning("Comparer failed due to type difference between {} and {}".format(
                type(b).__name__, type(self._cls).__name__
            ))
            return False

        comparer = self._get_comparer()
        return self.default(b) if comparer is None else comparer(self.object).equals(b)

    def default(self, b):
        return self._object == b

    def _get_comparer(self):
        # one can create on the fly subclasses, therefore we have to check it
        if len(self.__registry_cache) != len(self.registry):
            for cls in self.registry:
                if not hasattr(cls, 'type'):
                    raise TypeError('The subclass "{}" must have a "type" attribute'.format(cls.__name__))
                self.__registry_cache[cls.type] = cls

        # registry is updated
        # check if we can resolve it directly
        if self._cls in self.__registry_cache:
            return self.__registry_cache[self._cls]
        else:
            # it could be the subclass of one of the entries
            for k in self.__registry_cache.keys():
                if issubclass(self._cls, k):
                    return self.__registry_cache[k]

            return None

    def equals(self, b):
        return self.default(b)

    def __eq__(self, other):
        return self._equals(other)


class NumpyArrayComparer(Comparer):
    """
    Used to compare numpy arrays.
    """

    type = np.ndarray

    @staticmethod
    def get_machine_epsilon(a):
        """
        Returns the machine inaccuracy for the datatype of `a`.

        Args:
            a: (np.ndarray) the array

        Returns: (float or None) the machine epsilon or None if it is an exact datatype

        """
        try:
            epsilon = np.finfo(a.dtype).eps
        except ValueError:
            epsilon = None
        return epsilon

    def equals(self, b):
        fudge_factor = 10
        epsilon = self.get_machine_epsilon(b)
        # check if the datatype is inexact at all
        inexact = epsilon is not None
        if inexact:
            return self.object.shape == b.shape and np.allclose(self.object, b, atol=fudge_factor*epsilon, rtol=0)
        else:
            # it is an exact data type such as int
            return np.array_equal(self.object, b)


class AtomsComparer(Comparer):
    """
    Used to compare pyiron Atoms objects.
    """

    type = Atoms

    def equals(self, b):
        assert isinstance(b, Atoms)
        assert isinstance(self.object, Atoms)

        index_spec_mapping = lambda atoms: {site.index: site.symbol for site in atoms}
        # compare structures
        # https://github.com/pyiron/pyiron/blob/c447ffb4f1e003d0ebaced50a12def46beefab4f/pyiron/atomistics/job/interactive.py
        conditions = [
            len(self.object) == len(b),
            Comparer(self.object.cell.array) == b.cell.array,
            Comparer(self.object.get_scaled_positions()) == b.get_scaled_positions(),
            Comparer(self.object.get_initial_magnetic_moments()) == b.get_initial_magnetic_moments(),
            index_spec_mapping(self.object) == index_spec_mapping(b)
        ]
        return all(conditions)


class ListComparer(Comparer):
    """
    Used to compare lists.
    """

    type = list

    def equals(self, b):
        assert isinstance(b, list)
        assert isinstance(self.object, list)

        conditions = [
            len(self.object) == len(b),
            all([Comparer(val) == last_val for val, last_val in zip(self.object, b)])
        ]

        return all(conditions)



"""
Abstract base class for fitting interactomic potentials.
"""

from typing import List

import abc

import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.patheffects import withStroke
import numpy as np

from pyiron_base import FlattenedStorage
from pyiron_contrib.atomistics.atomistics.job.trainingcontainer import (
    TrainingContainer,
    TrainingStorage,
)


class PotentialFit(abc.ABC):
    """
    Abstract mixin that defines a general interface to potential fitting codes.

    Training data can be added to the job with :method:`~.add_training_data`.  This should be atom structures with
    (at least) corresponding energies and forces, but additional (per structure or atom) maybe added.  Subclasses of
    :class:`~.TrainingContainer` that define and handle such data are explicitly allowed.

    :property:`~.training_data` and :property:`~.predicted_data` can be used to access the initial training data and the
    predicted data on them after the fit.
    """

    @abc.abstractmethod
    def _add_training_data(self, container: TrainingContainer) -> None:
        pass

    def add_training_data(self, container: TrainingContainer) -> None:
        """
        Add data to the fit.

        Calling this multiple times appends data to internal storage.

        Args:
            container (:class:`.TrainingContainer`): container holding data to fit
        """
        if self.status.initialized:
            self._add_training_data(container)
        else:
            raise ValueError("Data can only be added before fitting is started!")

    @abc.abstractmethod
    def _get_training_data(self) -> TrainingStorage:
        pass

    @property
    def training_data(self) -> TrainingStorage:
        """
        Return all training data added so far.

        Returns:
            :class:`pyiron_contrib.atomistics.atomistics.job.trainingcontainer.TrainingStorage`: container holding all training data
        """
        return self._get_training_data()

    @abc.abstractmethod
    def _get_predicted_data(self) -> FlattenedStorage:
        pass

    @property
    def predicted_data(self) -> FlattenedStorage:
        """
        Predicted properties of the training data after the fit.

        In contrast to :property:`~.training_data` this may not contain the original atomic structures, but must be in
        the same order.  Certain properties in the training data may be omitted from this data set, if the inconvenient
        or impossible to predict.  This should be documented on the subclass for each specific code.

        Returns:
            :class:`pyiron_base.FlattenedStorage`: container holding all predictions of the fitted potential on the
                                                   training data
        """
        if self.status.finished:
            return self._get_predicted_data()
        else:
            raise ValueError("Data can only be accessed after successful fit!")

    @property
    def plot(self):
        """
        Plots correlation and (training) error histograms.
        """
        return PotentialPlots(self.training_data, self.predicted_data)

    @abc.abstractmethod
    def get_lammps_potential(self) -> pd.DataFrame:
        """
        Return a pyiron compatible dataframe that defines a potential to be used with a Lammps job (or subclass
        thereof).

        Returns:
            DataFrame: contains potential information to be used with a Lammps job.
        """
        pass


class PotentialPlots:
    def __init__(self, training_data, predicted_data):
        self._training_data = training_data
        self._predicted_data = predicted_data

    def energy_scatter_histogram(self):
        """
        Plots correlation and (training) error histograms.
        """
        energy_train = self._training_data["energy"] / self._training_data["length"]
        energy_pred = self._predicted_data["energy"] / self._predicted_data["length"]
        plt.subplot(1, 2, 1)
        plt.scatter(energy_train, energy_pred)

        plt.xlabel("True Energy Per Atom [eV / atom]")
        plt.ylabel("Predicted Energy Per Atom [eV / atom]")
        plt.plot()

        plt.subplot(1, 2, 2)
        plt.hist(energy_train - energy_pred)
        plt.xlabel("Training Error [eV / atom]")

    def energy_log_histogram(self, bins=20, logy=False):
        """
        Plots a histogram of logarithmic training errors.

        Bins are created automatically using the minimum and maximum absolute
        errors with the given number of bins.

        Arguments:
            bins (int, optional): number of bins for the histogram
            logy (bool, optional): if True use a log scale also for the y-axis
        """

        energy_train = self._training_data["energy"] / self._training_data["length"]
        energy_pred = self._predicted_data["energy"] / self._predicted_data["length"]
        de = abs(energy_train - energy_pred)
        rmse = np.sqrt((de**2).mean())
        mae  = de.mean()
        high = de.max()
        low  = de.min()

        ax = plt.gca()
        trafo = ax.get_xaxis_transform()
        def annotated_vline(x, text, linestyle='--'):
            plt.axvline(x, color='k', linestyle=linestyle)
            plt.text(
                    x=x, y=0.5, s=text,
                    transform=trafo,
                    rotation='vertical',
                    horizontalalignment='center',
                    path_effects=[withStroke(linewidth=4, foreground='w')],
            )

        plt.hist(
                de,
                bins=np.logspace(np.log10(low), np.log10(high), bins),
                log=logy
        )
        plt.xscale('log')
        annotated_vline(rmse, f'RMSE = {rmse:.02}')
        annotated_vline(mae,  f'MAE = {mae:.02}')
        annotated_vline(high, f'HIGH = {high:.02}', linestyle='-')
        annotated_vline(low,  f'LOW = {low:.02}', linestyle='-')
        plt.xlabel('Training Error [eV/atom]')

    def force_scatter_histogram(self, axis=None):
        """
        Plots correlation and (training) error histograms.

        Args:
            axis (None, int): Whether to plot for an axis or norm

        """
        force_train = self._training_data["forces"]
        force_pred = self._predicted_data["forces"]

        if axis is None:
            ft = np.linalg.norm(force_train, axis=1)
            fp = np.linalg.norm(force_pred, axis=1)
        else:
            ft = force_train[:, axis]
            fp = force_pred[:, axis]

        plt.subplot(1, 2, 1)
        plt.scatter(ft, fp)
        plt.xlabel("True Forces [eV/$\mathrm{\AA}$]")
        plt.ylabel("Predicted Forces [eV/$\AA$]")
        plt.subplot(1, 2, 2)
        plt.hist(ft - fp)
        plt.xlabel("Training Error [eV/$\AA$]")



from __future__ import annotations

import inspect
from functools import partialmethod
from typing import get_args, get_type_hints, Optional, TYPE_CHECKING

from pyiron_contrib.workflow.channels import (
    InputData, OutputData, InputSignal, OutputSignal
)
from pyiron_contrib.workflow.has_to_dict import HasToDict
from pyiron_contrib.workflow.io import Inputs, Outputs, Signals

if TYPE_CHECKING:
    from pyiron_contrib.workflow.workflow import Workflow


class Node(HasToDict):
    """
    Nodes have input and output channels that interface with the outside world, and
    a callable that determines what they actually compute. After running, their output
    channels are updated with the results of the node's computation, which
    triggers downstream node updates if those output channels are connected to other
    input channels.

    Nodes can be forced to run, or more gently "updated", which will trigger a
    calculation only if all of the input is ready.

    Nodes won't update themselves while setting inputs to initial values, but can
    optionally update themselves at the end instantiation.

    Nodes must be instantiated with a callable to deterimine their function, and a tuple
    of strings to name each returned value of that callable.

    The node label (unless otherwise provided), IO types, and input defaults for the
    _automatically_ from introspection of the node function.
    Additional properties like storage priority (present but doesn't do anything yet)
    and ontological type (not yet present) can be set using kwarg dictionaries with
    keys corresponding to the channel labels (i.e. the node arguments of the node
    function, or the output labels provided).

    Actual node instances can either be instances of the base node class, in which case
    the callable node function and output labels *must* be provided, in addition to
    other data, OR they can be instances of children of this class.
    Those children may define some or all of the node behaviour at the class level, and
    modify their signature accordingly so this is not available for alteration by the
    user, e.g. the node function and output labels may be hard-wired.

    Args:
        node_function (callable): The function determining the behaviour of the node.
        output_labels (tuple[str]): A name for each return value of the node function.
        label (str): The node's label. (Defaults to the node function's name.)
        run_automatically (bool): Whether to run when you are updated and all your
            input is ready. (Default is True).
        update_on_instantiation (bool): Whether to force an update at the end of instantiation.
            (Default is False.)
        **kwargs: Any additional keyword arguments whose keyword matches the label of an
            input channel will have their value assigned to that channel.

    Attributes:
        inputs (Inputs): A collection of input channels.
        outputs (Outputs): A collection of output channels.
        ready (bool): All input reports ready.
        connected (bool): Any IO channel has at least one connection.
        fully_connected (bool): Every IO channel has at least one connection.

    Methods:
        update: If `run_automatically` is true and all your input is ready, will
            run the engine.
        run: Parse and process the input, execute the engine, process the results and
            update the output.
        disconnect: Disconnect all IO connections.

    Note:
        The number of return values for the node function, and the number of output
        labels provided must be consistent!

    Examples:
        At the most basic level, to use nodes all we need to do is provide the `Node`
        class with a function and labels for its output, like so:
        >>> from pyiron_contrib.workflow.node import Node
        >>>
        >>> def mwe(x, y):
        ...     return x+1, y-1
        >>>
        >>> plus_minus_1 = Node(mwe, "p1", "m1")
        >>>
        >>> print(plus_minus_1.outputs.p1)
        None

        There is no output because we haven't given our function any input, and it has
        no defaults!
        However, we'll run into a hiccup if we try to update one of the inputs...
        >>> plus_minus_1.inputs.x = 1
        TypeError

        This is because updating an input value triggers the node to update -- i.e. it
        checks if all it's input is of a valid type, and if so attempts to execute its
        node function.
        In this case, our input is untyped -- so it's always considered valid -- and the
        type error comes from our `y - 1` term in the function, which is `None - 1`.

        There are three ways to resolve this: First, we could set
        `run_automatically = False`, then the node would not execute until we
        manually call the `run()` method.
        This impacts the long-term behaviour of the node though, so let's keep
        searching.

        We can provide initial values for our node function at instantiation using our
        kwargs.
        The node update is deferred until _all_ of these initial values are processed.
        Thus, the second solution is to ensure that _all_ the arguments of our function
        are receiving good enough initial values to facilitate an execution of the node
        function at the end of instantiation:
        >>> plus_minus_1 = Node(mwe, "p1", "m1", x=1, y=2)
        >>>
        >>> print(plus_minus_1.outputs.to_value_dict())
        {'p1': 2, 'm1': 1}

        Second, we could add type hints/defaults to our function so that it knows better
        than to try to evaluate itself with bad data.
        Let's make a new node following the second path.

        In this example, note the mixture of old-school (`typing.Union`) and new (`|`)
        type hints as well as nested hinting with a union-type inside the tuple for the
        return hint.
        Our treatment of type hints is **not infinitely robust**, but covers a wide
        variety of common use cases.
        >>> from typing import Union
        >>>
        >>> def hinted_example(
        ...     x: Union[int, float],
        ...     y: int | float = 1
        ... ) -> tuple[int, int | float]:
        ...     return x+1, y-1
        >>>
        >>> plus_minus_1 = Node(hinted_example, "p1", "m1")
        >>>
        >>> plus_minus_1.inputs.x = 1
        >>> print(plus_minus_1.outputs.to_value_dict())
        {'p1': 2, 'm1': 0}

        In this case, we're able to use the default value for `y`, but you can
        experiment with updating `y` first (when `x` still has the invalid value of
        `None`) to verify that the update is not triggered until _both_ inputs have
        valid values.

        In these examples, we've instantiated nodes directly from the base `Node` class,
        and populated their input directly with data.
        In practice, these nodes are meant to be part of complex workflows; that means
        both that you are likely to have particular nodes that get heavily re-used, and
        that you need the nodes to pass data to each other.

        For reusable nodes, we want to create a sub-class of `Node` that fixes some of
        the node behaviour -- usually the `node_function` and `output_labels`.
        There are two straightforward ways to accomplish this.
        The first is to override the `__init__` method directly:
        >>> from typing import Literal, Optional
        >>>
        >>> class AlphabetModThree(Node):
        ...     def __init__(
        ...         self,
        ...         label: Optional[str] = None,
        ...         input_storage_priority: Optional[dict[str, int]] = None,
        ...         output_storage_priority: Optional[dict[str, int]] = None,
        ...         run_automatically: bool = True,
        ...         update_on_instantiation: bool = False,
        ...         **kwargs
        ...     ):
        ...         super().__init__(
        ...             self.alphabet_mod_three,
        ...             "letter",
        ...             labe=label,
        ...             input_storage_priority=input_storage_priority,
        ...             output_storage_priority=output_storage_priority,
        ...             run_automatically=run_automatically,
        ...             update_on_instantiation=update_on_instantiation,
        ...             **kwargs
        ...         )
        ...
        ...     @staticmethod
        ...     def alphabet_mod_three(i: int) -> Literal["a", "b", "c"]:
        ...         return ["a", "b", "c"][i % 3]

        The second effectively does the same thing, but leverages python's
        `functools.partialmethod` to do so much more succinctly.
        In this example, note that the function is declared _before_ `__init__` is set,
        so that it is available in the correct scope (above, we could place it
        afterwards because we were accessing it through self).
        >>> from functools import partialmethod
        >>>
        >>> class Adder(Node):
        ...     @staticmethod
        ...     def adder(x: int = 0, y: int = 0) -> int:
        ...         return x + y
        ...
        ...     __init__ = partialmethod(
        ...         Node.__init__,
        ...         node_function=adder,
        ...         output_labels="sum",
        ...     )

        Finally, instead of setting input to a particular data value, we'll set it to
        be another node's output channel, thus forming a connection.
        When we update the upstream node, we'll see the result passed downstream:
        >>> adder = Adder()
        >>> alpha = AlphabetModThree(i=adder.outputs.sum)
        >>>
        >>> adder.inputs.x = 1
        >>> print(alpha.outputs.letter)
        "b"
        >>> adder.inputs.y = 1
        >>> print(alpha.outputs.letter)
        "c"
        >>> adder.inputs.x = 0
        >>> adder.inputs.y = 0
        >>> print(alpha.outputs.letter)
        "a"

        To see how to use many nodes together, look at the `Workflow` class.
    """
    def __init__(
            self,
            node_function: callable,
            *output_labels: str,
            label: Optional[str] = None,
            input_storage_priority: Optional[dict[str, int]] = None,
            output_storage_priority: Optional[dict[str, int]] = None,
            run_automatically: bool = True,
            update_on_instantiation: bool = True,
            workflow: Optional[Workflow] = None,
            **kwargs
    ):
        self.node_function = node_function
        self.label = label if label is not None else node_function.__name__

        self.workflow = None
        if workflow is not None:
            workflow.add(self)

        input_channels = self._build_input_channels(input_storage_priority)
        self.inputs = Inputs(*input_channels)

        output_channels = self._build_output_channels(
            *output_labels, storage_priority=output_storage_priority
        )
        self.outputs = Outputs(*output_channels)

        self.signals = self._build_signal_channels()

        self.run_automatically = False
        for k, v in kwargs.items():
            if k in self.inputs.labels:
                if isinstance(v, OutputData):
                    self.inputs[k] = v
                else:
                    self.inputs[k].update(v)
        self.run_automatically = run_automatically

        if update_on_instantiation:
            self.update()

    def _build_input_channels(self, storage_priority: dict[str:int]):
        channels = []
        type_hints = get_type_hints(self.node_function)
        parameters = inspect.signature(self.node_function).parameters

        for label, value in parameters.items():
            try:
                priority = storage_priority[label]
            except (KeyError, TypeError):
                priority = None

            try:
                type_hint = type_hints[label]
            except KeyError:
                type_hint = None

            if value.default is not inspect.Parameter.empty:
                default = value.default
            else:
                default = None

            channels.append(InputData(
                label=label,
                node=self,
                default=default,
                type_hint=type_hint,
                storage_priority=priority,
            ))
        return channels

    def _build_output_channels(
            self, *return_labels: str, storage_priority: dict[str:int] = None
    ):
        try:
            type_hints = get_type_hints(self.node_function)["return"]
            if len(return_labels) > 1:
                type_hints = get_args(type_hints)
                if not isinstance(type_hints, tuple):
                    raise TypeError(
                        f"With multiple return labels expected to get a tuple of type "
                        f"hints, but got type {type(type_hints)}"
                    )
                if len(type_hints) != len(return_labels):
                    raise ValueError(
                        f"Expected type hints and return labels to have matching "
                        f"lengths, but got {len(type_hints)} hints and "
                        f"{len(return_labels)} labels: {type_hints}, {return_labels}"
                    )
            else:
                # If there's only one hint, wrap it in a tuple so we can zip it with
                # *return_labels and iterate over both at once
                type_hints = (type_hints,)
        except KeyError:
            type_hints = [None] * len(return_labels)

        channels = []
        for label, hint in zip(return_labels, type_hints):
            try:
                priority = storage_priority[label]
            except (KeyError, TypeError):
                priority = None

            channels.append(
                OutputData(
                    label=label,
                    node=self,
                    type_hint=hint,
                    storage_priority=priority,
                )
            )

        return channels

    def _build_signal_channels(self):
        signals = Signals()
        signals.input.run = InputSignal("run", self, self.run)
        signals.output.ran = OutputSignal("ran", self)
        return signals

    def update(self) -> None:
        if self.run_automatically and self.ready:
            self.run()

    def run(self) -> None:
        function_output = self.node_function(**self.inputs.to_value_dict())

        if len(self.outputs) == 1:
            function_output = (function_output,)

        for out, value in zip(self.outputs, function_output):
            out.update(value)

        self.signals.output.ran()

    def __call__(self) -> None:
        self.run()

    def disconnect(self):
        self.inputs.disconnect()
        self.outputs.disconnect()
        self.signals.disconnect()

    @property
    def ready(self) -> bool:
        return self.inputs.ready

    @property
    def connected(self) -> bool:
        return self.inputs.connected or self.outputs.connected or self.signals.connected

    @property
    def fully_connected(self):
        return self.inputs.fully_connected \
            and self.outputs.fully_connected \
            and self.signals.fully_connected

    def set_storage_priority(self, priority: int):
        self.inputs.set_storage_priority(priority)
        self.outputs.set_storage_priority(priority)

    def to_dict(self):
        return {
            "label": self.label,
            "ready": self.ready,
            "connected": self.connected,
            "fully_connected": self.fully_connected,
            "inputs": self.inputs.to_dict(),
            "outputs": self.outputs.to_dict(),
            "signals": self.signals.to_dict()
        }


def node(*output_labels: str):
    """
    A decorator for dynamically creating node classes from functions.

    Decorates a function.
    Takes an output label for each returned value of the function.
    Returns a `Node` subclass whose name is the camel-case version of the function node,
    and whose signature is modified to exclude the node function and output labels
    (which are explicitly defined in the process of using the decorator).
    """

    def as_node(node_function: callable):
        return type(
            node_function.__name__.title().replace("_", ""),  # fnc_name to CamelCase
            (Node,),  # Define parentage
            {'__init__': partialmethod(Node.__init__, node_function, *output_labels)}
        )

    return as_node



import warnings

from pyiron_base import Toolkit
from pyiron_base import ImportAlarm

try:
    from pyiron_contrib.generic.coscineIo import CoscineProject, CoscineResource

    import_alarm = ImportAlarm()
except ImportError:
    import_alarm = ImportAlarm("Connecting to CoScInE requires the coscine package.")

    class CoscineProject:
        def __init__(self, *args, **kwargs):
            raise ImportError("Interacting with coscine requires the coscine package")

    class CoscineResource:
        def __init__(self, *args, **kwargs):
            raise ImportError("Interacting with coscine requires the coscine package")


from pyiron_contrib.generic.s3io import FileS3IO


class StorageInterfaceCreator:
    @import_alarm
    def __init__(self):
        pass

    @staticmethod
    def s3(*args, **kwargs):
        return FileS3IO(*args, **kwargs)

    @staticmethod
    def coscine(*args, **kwargs):
        return CoscineProject(*args, **kwargs)


class StorageInterfaceConnector:

    _known_storage_classes = {str(CoscineResource): "coscine", str(FileS3IO): "s3"}

    @import_alarm
    def __init__(self, project):
        self._store = {}
        if project is None:
            return

        if "StorageInterface" in project.data:
            self._data = project.data.StorageInterface.copy()
        else:
            self._data = {}

    @classmethod
    def from_dict(cls, info_dict):
        self = cls(None)
        name = info_dict.get("name", "new")
        self._data = {name: info_dict}
        self._connect_storage(name)

        return self

    def _connect_storage(self, name):
        info_dict = self._data[name]
        if info_dict["type"] == str(CoscineResource):
            self._store[name] = CoscineResource(info_dict["info"])
        elif info_dict["type"] == str(FileS3IO):
            self._store[name] = FileS3IO.from_dict(info_dict["info"])

    def connect_all_storages(self):
        for key in self._data:
            self._connect_storage(key)

    def attach_temporary(self, key, value, data_dict):
        self._store[key] = value
        self._data[key] = data_dict

    def __getitem__(self, item):
        if item not in self._store and item in self._data:
            self._connect_storage(item)
        elif item not in self._store:
            raise KeyError(item)
        return self._store[item]

    @property
    def info(self):
        result = {}
        for key in self._data:
            if key in self._store:
                result[key] = {"type": self._data[key]["type"], "connected": True}
            else:
                result[key] = {"type": self._data[key]["type"], "connected": False}
        return result

    def __repr__(self):
        result = []
        for key, value in self.info.items():
            conn = "connected" if value["connected"] else "inactive"
            storage_class = value["type"]
            storage_type = (
                self._known_storage_classes[storage_class]
                if storage_class in self._known_storage_classes
                else storage_class
            )
            result.append(f"{key}({storage_type}, {conn})")
        return f"Storage Access for {result}."


class StorageInterfaceFactory(Toolkit):
    def __init__(self, project):
        super().__init__(project)
        self._creator = None
        self._storage_interface = None

    @property
    def create(self):
        if self._creator is None:
            self._creator = StorageInterfaceCreator()
        return self._creator

    def attach(self, storage_name, storage_interface):
        info_dict = {
            "name": storage_name,
            "type": str(type(storage_interface)),
            "info": storage_interface.connection_info,
        }
        try:
            new = StorageInterfaceConnector.from_dict(info_dict)
        except ValueError as e:
            raise ValueError(
                "Credential information insufficient to auto-connect - storage interface not saved!"
            ) from e
        else:
            if "StorageInterface" not in self._project.data:
                self._project.data.create_group("StorageInterface")
            self._project.data.StorageInterface[storage_name] = info_dict
            self._project.data.write()
            if self._storage_interface is None:
                self._storage_interface = new
            else:
                self._storage_interface.attach_temporary(
                    storage_name, new[storage_name], info_dict
                )

    @property
    def storage(self):
        if self._storage_interface is None:
            self._storage_interface = StorageInterfaceConnector(self._project)
        return self._storage_interface



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_base import JobStatus, GenericJob, GenericParameters
from pyiron.atomistics.job.interactivewrapper import InteractiveWrapper, ReferenceJobOutput

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class Mixer(InteractiveWrapper):
    """
    Mixing forces of two atomistic interpreters, e.g. a LAMMPS or VASP job

    Args:
        project (ProjectHDFio): ProjectHDFio instance which points to the HDF5 file the job is stored in
        job_name (str): name of the job, which has to be unique within the project

    Attributes:
        input (pyiron_base.GenericParameters): handles the input
        ref_job_0 (pyiron.atomistics.job.atomistic.AtomisticGenericJob): First atomistic interpreter
        ref_job_1 (pyiron.atomistics.job.atomistic.AtomisticGenericJob): Second atomistic interpreter
    """

    def __init__(self, project, job_name):
        super(Mixer, self).__init__(project, job_name)
        self.__name__ = "Mixer"
        self.input = Input()
        self._ref_job_0 = None
        self._ref_job_1 = None
        self.output = MixingOutput(job=self)
        self.interactive_cache = {}
        self.server.run_mode.interactive = True

    @property
    def ref_job(self):
        """
        Compatibility function - it is recommended to use ref_job_0 instead

        Returns:
            pyiron.atomistics.job.atomistic.AtomisticGenericJob: Reference Job
        """
        return self.ref_job_0

    @ref_job.setter
    def ref_job(self, ref_job):
        """
        Compatibility function - it is recommended to use ref_job_0 instead

        Args:
            ref_job (pyiron.atomistics.job.atomistic.AtomisticGenericJob): Reference Job
        """
        self.ref_job_0 = ref_job

    @property
    def ref_job_0(self):
        """
        First Quantum engine to mix forces

        Returns:
            pyiron.atomistics.job.atomistic.AtomisticGenericJob: Reference Job
        """
        if self._ref_job_0 is not None:
            return self._ref_job_0
        try:
            if isinstance(self[0], GenericJob):
                self._ref_job_0 = self[0]
                self._ref_job_0._job_id = None
                self._ref_job_0._status = JobStatus(db=self.project.db)
                return self._ref_job_0
            else:
                return None
        except IndexError:
            return None

    @ref_job_0.setter
    def ref_job_0(self, ref_job):
        """
        First Quantum engine to mix forces

        Args:
            ref_job (pyiron.atomistics.job.atomistic.AtomisticGenericJob): Reference Job
        """
        self.append(ref_job)

    @property
    def ref_job_1(self):
        """
        Second Quantum engine to mix forces

        Returns:
            pyiron.atomistics.job.atomistic.AtomisticGenericJob: Reference Job
        """
        if self._ref_job_1 is not None:
            return self._ref_job_1
        try:
            if isinstance(self[1], GenericJob):
                self._ref_job_1 = self[1]
                self._ref_job_1._job_id = None
                self._ref_job_1._status = JobStatus(db=self.project.db)
                return self._ref_job_1
            else:
                return None
        except IndexError:
            return None

    @ref_job_1.setter
    def ref_job_1(self, ref_job):
        """
        Second Quantum engine to mix forces

        Args:
            ref_job (pyiron.atomistics.job.atomistic.AtomisticGenericJob): Reference Job
        """
        if self.ref_job_0 is None:
            raise ValueError('Please assign ref_job_0 before ref_job_1.')
        ref_job.structure = self._ref_job_0.structure
        self.append(ref_job)

    @property
    def structure(self):
        """
        Atomistic Structure - returns structure of first reference job

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: Atomistic crystal structure
        """
        if self.ref_job is not None:
            return self._ref_job_0.structure
        else:
            return None

    @structure.setter
    def structure(self, basis):
        """
        Atomistic Structure - set structure of first reference job

        Args:
            basis (pyiron.atomistics.structure.atoms.Atoms): Atomistic crystal structure
        """
        if self.ref_job_0 is not None and self.ref_job_1 is not None:
            self._ref_job_0.structure = basis
            self._ref_job_1.structure = basis
        else:
            raise ValueError('A structure can only be set after both ref_jobs (ref_job_0 and ref_job_1) have been' 
                             'assinged.')

    def set_input_to_read_only(self):
        """
        This function enforces read-only mode for the input classes, but it has to be implement in the individual
        classes.
        """
        self.input.read_only = True

    def write_input(self):
        """
        No input is written for the Mixer class
        """
        pass

    def _write_run_wrapper(self, debug=False):
        """
        No wrapper is writen for the Mixer class

        Args:
            debug (bool): the debug flag is ignored
        """
        pass

    def interactive_initialize_interface(self):
        """
        Internal helper function to initialize the interactive interface
        """
        for key in list(set(list(self.ref_job_0.interactive_cache.keys()) +
                            list(self.ref_job_1.interactive_cache.keys()))):
            self.interactive_cache[key] = []
        self.output = MixingOutput(job=self)

    def run_if_interactive(self):
        """
        The interactive run function calls run on both quantum engines - both interactive and non-interactive jobs
        are supported.
        """
        self.status.running = True
        if self.ref_job_0.server.run_mode.interactive:
            self.ref_job_0.run()
        else:
            self.ref_job_0.run(run_again=True)
        if self.ref_job_1.server.run_mode.interactive:
            self.ref_job_1.run()
        else:
            self.ref_job_1.run(run_again=True)

    def interactive_store_in_cache(self, key, value):
        """
        Helper function to store the output in the cache - storing the same information in both reference jobs

        Args:
            key (str): Key in cache
            value (list/float/int): value to store
        """
        self.ref_job_0.interactive_cache[key] = value
        self.ref_job_1.interactive_cache[key] = value

    def interactive_close(self):
        """
        Internal helper function to close the interactive interface.
        """
        self.status.collect = True
        if self.ref_job_0.server.run_mode.interactive:
            self.ref_job_0.interactive_close()
        if self.ref_job_1.server.run_mode.interactive:
            self.ref_job_1.interactive_close()
        self.project.db.item_update(self._runtime(), self.job_id)
        self.status.finished = True

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore the Mixer from an HDF5 file

        Args:
            hdf (ProjectHDFio): HDF5 group object - optional
            group_name (str): HDF5 subgroup name - optional
        """
        super(Mixer, self).from_hdf(hdf=hdf, group_name=group_name)
        self.interactive_initialize_interface()


class Input(GenericParameters):
    """
    class to control the generic input for a Mixer calculation.

    Args:
        input_file_name (str): name of the input file
        table_name (str): name of the GenericParameters table
    """

    def __init__(self, input_file_name=None, table_name="input"):
        super(Input, self).__init__(input_file_name=input_file_name, table_name=table_name, comment_char="//",
                                    separator_char="=", end_value_char=';')

    def load_default(self):
        """
        Loads the default file content
        """
        file_content = ('lambda = 0.5\n')
        self.load_string(file_content)


class MixingOutput(ReferenceJobOutput):
    def __init__(self, job):
        """
        Output class to aggregate the combined output of both calculation, only in the output the energies and forces
        are mixed.

        Args:
            job (Mixer): Corresponding Mixer job
        """
        super(MixingOutput, self).__init__(job=job)
        self._lambda = float(job.input['lambda'])

    @property
    def energy_pot(self):
        """
        Combined potential energy

        Returns:
            list: potential energy
        """
        return self._lambda * np.array(self._job.ref_job_0.output.energy_pot) + \
               (1-self._lambda) * np.array(self._job.ref_job_1.output.energy_pot)

    @property
    def energy_tot(self):
        """
        Combined total energy

        Returns:
            list: total energy
        """
        return self._lambda * np.array(self._job.ref_job_0.output.energy_tot) + \
               (1 - self._lambda) * np.array(self._job.ref_job_1.output.energy_tot)

    @property
    def forces(self):
        """
        Combined forces

        Returns:
            list: forces
        """
        return self._lambda * np.array(self._job.ref_job_0.output.forces) + \
               (1 - self._lambda) * np.array(self._job.ref_job_1.output.forces)

    @property
    def pressures(self):
        """
        Combined pressure

        Returns:
            list: pressure
        """
        return self._lambda * np.array(self._job.ref_job_0.output.pressures) + \
               (1 - self._lambda) * np.array(self._job.ref_job_1.output.pressures)

    @property
    def temperature(self):
        """
        Combined temperature

        Returns:
            list: temperature
        """
        return self._lambda * np.array(self._job.ref_job_0.output.temperature) + \
               (1 - self._lambda) * np.array(self._job.ref_job_1.output.temperature)



import pandas
import numpy as np
from pyiron_base import GenericMaster, state
from pyiron_atomistics.atomistics.master.murnaghan import Murnaghan


class MurnaghanWithoutFiles(Murnaghan):
    def __init__(self, project, job_name):
        if not state.database.database_is_disabled:
            raise RuntimeError(
                "To run a `Without` job, the database must first be disabled. Please "
                "`from pyiron_base import state; "
                "state.update({'disable_database': True})`, and try again."
            )
        super(MurnaghanWithoutFiles, self).__init__(project, job_name)
        self._interactive_disable_log_file = True

    @property
    def child_project(self):
        """
        :class:`.Project`: project which holds the created child jobs
        """
        if not self._interactive_disable_log_file:
            return super(MurnaghanWithoutFiles, self).child_project
        else:
            return self.project

    def to_hdf(self, hdf=None, group_name=None):
        """
        Args:
            hdf:
            group_name:
        Returns:
        """
        if not self._interactive_disable_log_file:
            super(MurnaghanWithoutFiles, self).to_hdf(hdf=hdf, group_name=group_name)

    def refresh_job_status(self):
        if not self._interactive_disable_log_file:
            super(MurnaghanWithoutFiles).refresh_job_status()

    def _store_fit_in_hdf(self, fit_dict):
        # implemented in https://github.com/pyiron/pyiron_atomistics/pull/960
        with self.project_hdf5.open("input") as hdf5_input:
            self.input.to_hdf(hdf5_input)
        with self.project_hdf5.open("output") as hdf5:
            hdf5["equilibrium_energy"] = fit_dict["energy_eq"]
            hdf5["equilibrium_volume"] = fit_dict["volume_eq"]
            hdf5["equilibrium_bulk_modulus"] = fit_dict["bulkmodul_eq"]
            hdf5["equilibrium_b_prime"] = fit_dict["b_prime_eq"]
        self._final_struct_to_hdf()

    def _fit_eos_general(self, vol_erg_dic=None, fittype="birchmurnaghan"):
        self._set_fit_module(vol_erg_dic=vol_erg_dic)
        fit_dict = self.fit_module.fit_eos_general(fittype=fittype)
        self.input["fit_type"] = fit_dict["fit_type"]
        self.input["fit_order"] = 0
        if not self._interactive_disable_log_file:
            self._store_fit_in_hdf(fit_dict=fit_dict)
        self.fit_dict = fit_dict
        return fit_dict

    def poly_fit(self, fit_order=3, vol_erg_dic=None):
        self._set_fit_module(vol_erg_dic=vol_erg_dic)
        fit_dict = self.fit_module.fit_polynomial(fit_order=fit_order)
        if fit_dict is None:
            self._logger.warning("Minimum could not be found!")
        elif not self._interactive_disable_log_file:
            self.input["fit_type"] = fit_dict["fit_type"]
            self.input["fit_order"] = fit_dict["fit_order"]
            if not self._interactive_disable_log_file:
                self._store_fit_in_hdf(fit_dict=fit_dict)
        self.fit_dict = fit_dict
        return fit_dict

    def collect_output(self):
        if not self._interactive_disable_log_file:
            super(MurnaghanWithoutFiles).collect_output()
        elif self.ref_job.server.run_mode.interactive:
            erg_lst = self.ref_job.output.energy_pot.copy()
            vol_lst = self.ref_job.output.volume.copy()
            arg_lst = np.argsort(vol_lst)

            self._output["volume"] = vol_lst[arg_lst]
            self._output["energy"] = erg_lst[arg_lst]
            if self.input["fit_type"] == "polynomial":
                self.fit_polynomial(fit_order=self.input["fit_order"])
            else:
                self._fit_eos_general(fittype=self.input["fit_type"])
        else:
            raise ValueError("No files execution requires interactive jobs.")

    def _run_if_collect(self):
        """
        Internal helper function the run if collect function is called when the job status is 'collect'. It collects
        the simulation output using the standardized functions collect_output() and collect_logfiles(). Afterwards the
        status is set to 'finished'.
        """

        if not self._interactive_disable_log_file:
            super(MurnaghanWithoutFiles)._run_if_collect()
        else:
            self._logger.info(
                "{}, status: {}, finished".format(self.job_info_str, self.status)
            )
            self.collect_output()
            self._logger.info(
                "{}, status: {}, parallel master".format(self.job_info_str, self.status)
            )
            self.update_master()
            self.status.finished = True
            # self.send_to_database()

    def append(self, job):
        """
        Append a job to the GenericMaster - just like you would append an element to a list.
        Args:
            job (GenericJob): job to append
        """
        if self.status.initialized and not job.status.initialized:
            raise ValueError(
                "GenericMaster requires reference jobs to have status initialized, rather than ",
                job.status.string,
            )
        if job.server.cores >= self.server.cores:
            self.server.cores = job.server.cores
        if job.job_name not in self._job_name_lst:
            self._job_name_lst.append(job.job_name)
            if not self._interactive_disable_log_file:
                self._child_job_update_hdf(parent_job=self, child_job=job)

    def pop(self, i=-1):
        """
        Pop a job from the GenericMaster - just like you would pop an element from a list
        Args:
            i (int): position of the job. (Default is last element, -1.)
        Returns:
            GenericJob: job
        """
        job_name_to_return = self._job_name_lst[i]
        job_to_return = self._load_all_child_jobs(
            self._load_job_from_cache(job_name_to_return)
        )
        del self._job_name_lst[i]
        if not self._interactive_disable_log_file:
            with self.project_hdf5.open("input") as hdf5_input:
                hdf5_input["job_list"] = self._job_name_lst
            job_to_return.relocate_hdf5()
        if isinstance(job_to_return, GenericMaster):
            for sub_job in job_to_return._job_object_dict.values():
                self._child_job_update_hdf(parent_job=job_to_return, child_job=sub_job)
        job_to_return.status.initialized = True
        return job_to_return

    def output_to_pandas(self, sort_by=None, h5_path="output"):
        """
        Convert output of all child jobs to a pandas Dataframe object.

        Args:
            sort_by (str): sort the output using pandas.DataFrame.sort_values(by=sort_by)
            h5_path (str): select child output to include - default='output'

        Returns:
            pandas.Dataframe: output as dataframe
        """
        # TODO: The output to pandas function should no longer be required
        if not self._interactive_disable_log_file:
            super(MurnaghanWithoutFiles, self).output_to_pandas(
                sort_by=sort_by, h5_path=h5_path
            )
        else:
            df = pandas.DataFrame(self._output)
            if sort_by is not None:
                df = df.sort_values(by=sort_by)
            return df



from pyiron_contrib.atomistics.atomicrex.interactive import AtomicrexInteractive


class Atomicrex(AtomicrexInteractive):
    """Class to set up and run atomicrex jobs"""

    def __init__(self, project, job_name):
        super().__init__(project, job_name)
        self.__name__ = "Atomicrex"
        self._executable_activate(enforce=True)
